#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A state of not fully resolved software stack in adviser's recommendations implementation."""

import math
from typing import Any
from typing import Tuple
from typing import Dict
from typing import List
from typing import Optional
from typing import Generator
import random
import weakref

import attr
from thoth.common import RuntimeEnvironment
from thoth.python import PackageVersion


@attr.s(slots=True, order=False)
class State:
    """Implementation of an adviser's state in state space."""

    score = attr.ib(type=float, default=0.0)
    # Iteration in which the state was introduced.
    iteration = attr.ib(type=int, default=0)
    # States added in the given iteration.
    unresolved_dependencies = attr.ib(default=attr.Factory(dict))  # type: Dict[str, Dict[int, Tuple[str, str, str]]]
    resolved_dependencies = attr.ib(default=attr.Factory(dict))  # type: Dict[str, Tuple[str, str, str]]
    _parent = attr.ib(default=None)  # type: weakref.ReferenceType['State']
    advised_runtime_environment = attr.ib(type=Optional[RuntimeEnvironment], kw_only=True, default=None)
    # See RFC-6902:
    #   https://tools.ietf.org/html/rfc6902
    # An example entry for a manifest change:
    #   [
    #     {
    #       "apiVersion:": "apps.openshift.io/v1",
    #       "kind": "DeploymentConfig",
    #       "patch": {
    #         "op": "add",
    #         "path": "spec.template.spec.containers[0].env",
    #         "value": {"name": "OMP_NUM_THREADS", "value": "1"},
    #       }
    #     }
    #   ]
    # See available yaml-patch operations at https://python-json-patch.readthedocs.io/en/latest/tutorial.html
    #
    # Each entry is a list of JSON patches to be applied. If any entry fails to
    # patch, the patching can continue with the subsequent JSON patch as each
    # entry corresponds to a list of atomic changes to be applied:
    #
    #  https://tools.ietf.org/html/rfc6902#section-5
    advised_manifest_changes = attr.ib(type=List[List[Dict[str, Any]]], kw_only=True, default=attr.Factory(list))
    justification = attr.ib(type=List[Dict[str, str]], default=attr.Factory(list), kw_only=True)

    _EPSILON = 0.1

    @property
    def parent(self):
        # type:('State') -> Optional['State']
        """Retrieve parent to this state.

        If None the state is top level state or parent is no longer maintained. Note the return
        value of None depends on actual gc runs.
        """
        if self._parent:
            return self._parent()

        return None

    @classmethod
    def from_direct_dependencies(cls, direct_dependencies: Dict[str, List[PackageVersion]]) -> "State":
        """Create an initial state out of direct dependencies."""
        unresolved_dependencies = {}  # type: Dict[str, Dict[int, Tuple[str, str, str]]]

        for dependency_name, dependency_versions in direct_dependencies.items():
            unresolved_dependencies[dependency_name] = {}
            for dependency_version in dependency_versions:
                dependency_tuple = dependency_version.to_tuple()
                unresolved_dependencies[dependency_name][hash(dependency_tuple)] = dependency_tuple

        return cls(unresolved_dependencies=unresolved_dependencies)

    def __lt__(self, other: "State") -> bool:
        """Compare two objects, comparision protocol used in the beam."""
        return self.score < other.score

    def to_dict(self) -> Dict[str, Any]:
        """Convert state to a dict representation."""
        advised_runtime_environment = None
        if self.advised_runtime_environment:
            advised_runtime_environment = self.advised_runtime_environment.to_dict()

        # Parent is removed intentionally.
        return {
            "advised_manifest_changes": self.advised_manifest_changes,
            "advised_runtime_environment": advised_runtime_environment,
            "iteration": self.iteration,
            "justification": self.justification,
            "resolved_dependencies": self.resolved_dependencies,
            "score": self.score,
            "unresolved_dependencies": self.unresolved_dependencies,
        }

    def is_final(self) -> bool:
        """Check if the given state is a final state."""
        return len(self.unresolved_dependencies) == 0

    def add_justification(self, justification: List[Dict[str, str]]) -> None:
        """Add new entries to the justification field."""
        self.justification.extend(justification)

    def add_unresolved_dependency(self, package_tuple: Tuple[str, str, str]) -> None:
        """Add unresolved dependency into the state."""
        if package_tuple[0] not in self.unresolved_dependencies:
            self.unresolved_dependencies[package_tuple[0]] = {}

        self.unresolved_dependencies[package_tuple[0]][hash(package_tuple)] = package_tuple

    def set_unresolved_dependencies(self, dependencies: Dict[str, List[Tuple[str, str, str]]]) -> None:
        """Set unresolved dependencies - any unresolved dependencies will be overwritten."""
        for dependency_name, dependency_tuples in dependencies.items():
            self.unresolved_dependencies[dependency_name] = {hash(d): d for d in dependency_tuples}

    def update_unresolved_dependencies(self, dependencies: Dict[str, List[Tuple[str, str, str]]]) -> None:
        """Update unresolved dependencies respecting the ones passed in as parameters."""
        for dependency_name, dependency_tuples in dependencies.items():
            if not dependency_tuples:
                continue

            if dependency_name not in self.unresolved_dependencies:
                self.unresolved_dependencies[dependency_name] = {}

            for d in dependency_tuples:
                self.unresolved_dependencies[dependency_name][hash(d)] = d

    def remove_unresolved_dependency(self, package_tuple: Tuple[str, str, str]) -> None:
        """Remove the given unresolved dependency from state."""
        self.unresolved_dependencies[package_tuple[0]].pop(hash(package_tuple))
        if not self.unresolved_dependencies[package_tuple[0]]:
            # Last item, remove records about it.
            self.unresolved_dependencies.pop(package_tuple[0])

    def remove_unresolved_dependency_subtree(self, package_name: str) -> None:
        """Remove the whole dependency sub-tree from the state."""
        self.unresolved_dependencies.pop(package_name, None)

    def add_resolved_dependency(self, package_tuple: Tuple[str, str, str]) -> None:
        """Add a resolved dependency into the state."""
        if (
            package_tuple[0] in self.resolved_dependencies
            and package_tuple != self.resolved_dependencies[package_tuple[0]]
        ):
            raise ValueError(
                f"Package {package_tuple!r} is already present in the state "
                f"in different version {self.resolved_dependencies[package_tuple[0]]!r}"
            )
        self.resolved_dependencies[package_tuple[0]] = package_tuple

    def mark_dependency_resolved(self, package_tuple: Tuple[str, str, str]) -> None:
        """Mark the given dependency as resolved in the current state."""
        self.remove_unresolved_dependency(package_tuple)
        self.remove_unresolved_dependency_subtree(package_tuple[0])
        # The logic behind state manipulation makes sure there is no package version or package source clash.
        self.add_resolved_dependency(package_tuple)

    def get_first_unresolved_dependency(self, dependency_name: Optional[str] = None) -> Tuple[str, str, str]:
        """Get a very first unresolved dependency tuple."""
        try:
            dependency_name = dependency_name or next(iter(self.unresolved_dependencies))
            unresolved_dependency_id = next(iter(self.unresolved_dependencies[dependency_name]))
            return self.unresolved_dependencies[dependency_name][unresolved_dependency_id]
        except StopIteration as exc:
            raise ValueError(f"No unresolved dependency found in state: {self!r}") from exc

    def get_random_first_unresolved_dependency(self, dependency_name: Optional[str] = None) -> Tuple[str, str, str]:
        """Get a very first unresolved dependency tuple."""
        dependency_name = dependency_name or random.choice(list(self.unresolved_dependencies))
        try:
            unresolved_dependency_id = next(iter(self.unresolved_dependencies[dependency_name]))
            return self.unresolved_dependencies[dependency_name][unresolved_dependency_id]
        except StopIteration as exc:
            raise ValueError(f"No unresolved dependency found in state: {self!r}") from exc

    @staticmethod
    def _termial_function(n: int) -> int:
        r"""Compute termial function, with red hats off to Donald Knuth.

        Termial function is:

          \sum_{k=0}^n k

        Let's use arithmetic progression to get termial of n.
        """
        if n < 0:
            raise ValueError("Termial function used with negative number")

        return int((1 + n) * (n / 2))

    @staticmethod
    def _termial_function_solution(x: int) -> int:
        """Compute solution to termial function.

        Let termial function be:

           (n**2 + n) / 2 = x

        This function computes non-negative ceil(n) so that it satisfies x.

        The solution is optimized to compute a generic parabolic intersection for termial function:

           x1 = (-b + sqrt(b**2 - 4*a*c)) / 2*a
           x2 = (-b - sqrt(b**2 - 4*a*c)) / 2*a

        Note solution x2 can be discarded based on termial function usage in the sources.

        The original formula:

           a*x**2 + b*x + c = 0

        for solution:

           (n**2)/2 + n/2 = x

        where:

          a = 1/2
          b = 1/2
          c = -x
        """
        return math.floor(-0.5 + math.sqrt(0.25 + (x << 1)))

    @classmethod
    def _random_termial(cls, n: int) -> int:
        """Compute a random number x such as 0 <= x < n, use termial function not to spread numbers uniformly.

        To prefer lower numbers more often, termial function is used to assign "weights" for
        numbers - then random uniform is used in conjunction with weights.

        An illustrative example can be cls._random_termial(4) which calls:

          cls._termial_function(4) == 10

        This way, we create 10 "cells" where (assignment range is inclusive):

           * 0 - 3 are assigned to number 3
           * 4 - 6 are assigned to number 2
           * 7 - 8 are assigned to number 1
           * 9 is assigned to number 0

        Now a random uniform call picks from 0 - 9 (inclusively), then we check what interval we have hit.
        The value returned inverts priority, so for cls._termial_function(4) probabilities are:

          * 0 - 4/10 = 0.4
          * 1 - 3/10 = 0.3
          * 2 - 2/10 = 0.2
          * 3 - 1/10 = 0.1
        """
        x = cls._termial_function(n)
        return n - 1 - cls._termial_function_solution(random.randrange(0, x))

    def get_random_unresolved_dependency(
        self, dependency_name: Optional[str] = None, prefer_recent: bool = True
    ) -> Tuple[str, str, str]:
        """Get a very first unresolved dependency tuple."""
        dependency_name = dependency_name or random.choice(list(self.unresolved_dependencies))

        choices = list(self.unresolved_dependencies[dependency_name])
        if prefer_recent:
            # perform multi-armed bandit - epsilon-greedy strategy
            unresolved_dependency_id = None
            if len(choices) > 1 and self._EPSILON >= random.random():
                unresolved_dependency_id = choices[self._random_termial(len(choices))]

            if unresolved_dependency_id is None:
                unresolved_dependency_id = choices[0]
        else:
            unresolved_dependency_id = random.choice(choices)

        return self.unresolved_dependencies[dependency_name][unresolved_dependency_id]

    def iter_unresolved_dependencies(self) -> Generator[Tuple[str, str, str], None, None]:
        """Iterate over unresolved dependencies."""
        for nested in self.unresolved_dependencies.values():
            yield from nested.values()

    def iter_resolved_dependencies(self) -> Generator[Tuple[str, str, str], None, None]:
        """Iterate over resolved dependencies."""
        yield from self.resolved_dependencies.values()

    def clone(self) -> "State":
        """Return a swallow copy of this state that can be used as a next state."""
        cloned_advised_environment = None
        if self.advised_runtime_environment:
            cloned_advised_environment = RuntimeEnvironment.from_dict(self.advised_runtime_environment.to_dict())

        unresolved_dependencies = self.unresolved_dependencies.copy()
        for dependency_name in unresolved_dependencies.keys():
            unresolved_dependencies[dependency_name] = unresolved_dependencies[dependency_name].copy()

        return self.__class__(
            score=self.score,
            iteration=self.iteration,
            unresolved_dependencies=unresolved_dependencies,
            resolved_dependencies=self.resolved_dependencies.copy(),
            advised_runtime_environment=cloned_advised_environment,
            advised_manifest_changes=self.advised_manifest_changes.copy(),
            justification=self.justification.copy(),
            parent=weakref.ref(self),
        )

    def __del__(self) -> None:
        """Destruct self."""
        # Destruct parts that are not eventually populated to the pipeline product abstraction.
        del self.unresolved_dependencies
        del self.resolved_dependencies



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test implementation of random walk in the dependency graph."""

from typing import Callable

import flexmock
from hypothesis import given
from hypothesis.strategies import integers

import random

from thoth.adviser.beam import Beam
from thoth.adviser.predictors import RandomWalk
from thoth.adviser.state import State

from ..base import AdviserTestCase


class TestRandomWalk(AdviserTestCase):
    """Tests related to random walk in the dependency graph."""

    @given(integers(min_value=1, max_value=256),)
    def test_run(self, state_factory: Callable[[], State], state_count: int) -> None:
        """Test running the random walk method."""
        state = state_factory()
        beam = Beam()
        for _ in range(state_count):
            cloned_state = state.clone()
            cloned_state.score = random.random()
            cloned_state.iteration = state.iteration + 1
            beam.add_state(cloned_state)

        predictor = RandomWalk()
        context = flexmock(accepted_final_states_count=33, beam=beam)

        with predictor.assigned_context(context):
            next_state, package_tuple = predictor.run()
            assert next_state in beam.iter_states()
            assert package_tuple is not None
            assert package_tuple[0] in next_state.unresolved_dependencies
            assert package_tuple in next_state.unresolved_dependencies[package_tuple[0]].values()

    def test_pre_run(self) -> None:
        """Test pre-run initialization."""
        context = flexmock(limit=99)

        predictor = RandomWalk()
        assert predictor._history == []
        predictor._history = [(0.99, 33), (0.42, 42)]

        with predictor.assigned_context(context):
            predictor.pre_run()
            assert predictor._history == [], "Predictor's history not discarded"



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test manipulation with a pipeline product."""


import flexmock

from thoth.adviser.product import Product
from thoth.adviser.state import State
from thoth.adviser.context import Context
from thoth.common import RuntimeEnvironment
from thoth.python import PackageVersion
from thoth.python import Pipfile
from thoth.python import Source

from .base import AdviserTestCase


class TestProduct(AdviserTestCase):
    """Test manipulation with product."""

    _PIPFILE = """
[[source]]
name = "pypi-org"
url = "https://pypi.org/simple"
verify_ssl = true

[dev-packages]

[packages]
tensorflow = "*"

[requires]
python_version = "3.7"
"""

    def test_from_final_state(self, context: Context) -> None:
        """Test instantiating product from a final state."""
        state = State(
            score=0.5,
            resolved_dependencies={
                "daiquiri": ("daiquiri", "1.6.0", "https://pypi.org/simple"),
                "numpy": ("numpy", "1.17.4", "https://pypi.org/simple"),
                "tensorflow": ("tensorflow", "2.0.0", "https://pypi.org/simple"),
            },
            unresolved_dependencies={},
            advised_runtime_environment=RuntimeEnvironment.from_dict({"python_version": "3.6"}),
        )
        state.add_justification(self.JUSTIFICATION_SAMPLE_1)

        pypi = Source("https://pypi.org/simple")

        pv_daiquiri_locked = PackageVersion(name="daiquiri", version="==1.6.0", index=pypi, develop=False)
        pv_numpy_locked = PackageVersion(name="numpy", version="==1.17.4", index=pypi, develop=False)
        pv_tensorflow_locked = PackageVersion(name="tensorflow", version="==2.0.0", index=pypi, develop=False)

        context.should_receive("get_package_version").with_args(
            ("daiquiri", "1.6.0", "https://pypi.org/simple"), graceful=False
        ).and_return(pv_daiquiri_locked).ordered()
        context.graph.should_receive("get_python_package_hashes_sha256").with_args(
            "daiquiri", "1.6.0", "https://pypi.org/simple"
        ).and_return(["000"]).ordered()

        context.should_receive("get_package_version").with_args(
            ("numpy", "1.17.4", "https://pypi.org/simple"), graceful=False
        ).and_return(pv_numpy_locked).ordered()
        context.graph.should_receive("get_python_package_hashes_sha256").with_args(
            "numpy", "1.17.4", "https://pypi.org/simple"
        ).and_return(["111"]).ordered()

        context.should_receive("get_package_version").with_args(
            ("tensorflow", "2.0.0", "https://pypi.org/simple"), graceful=False
        ).and_return(pv_tensorflow_locked).ordered()
        context.graph.should_receive("get_python_package_hashes_sha256").with_args(
            "tensorflow", "2.0.0", "https://pypi.org/simple"
        ).and_return(["222"]).ordered()

        pv_daiquiri = PackageVersion(name="daiquiri", version="*", index=pypi, develop=False)
        pv_tensorflow = PackageVersion(name="tensorflow", version=">=2.0.0", index=pypi, develop=False)

        project = flexmock(
            pipfile=Pipfile.from_string(self._PIPFILE),
            runtime_environment=RuntimeEnvironment.from_dict({"operating_system": {"name": "rhel"}}),
        )
        project.should_receive("iter_dependencies").with_args(with_devel=True).and_return(
            [pv_daiquiri, pv_tensorflow]
        ).once()

        context.project = project
        context.dependencies = {
            "daiquiri": {("daiquiri", "1.6.0", "https://pypi.org/simple"): set(),},
            "numpy": {("numpy", "1.17.4", "https://pypi.org/simple"): set()},
            "tensorflow": {
                ("tensorflow", "2.0.0", "https://pypi.org/simple"): {("numpy", "1.17.4", "https://pypi.org/simple")}
            },
        }
        context.dependents = {
            "daiquiri": {("daiquiri", "1.6.0", "https://pypi.org/simple"): set(),},
            "numpy": {
                ("numpy", "1.17.4", "https://pypi.org/simple"): {
                    (("tensorflow", "2.0.0", "https://pypi.org/simple"), "fedora", "31", "3.7",)
                }
            },
            "tensorflow": {("tensorflow", "2.0.0", "https://pypi.org/simple"): set()},
        }
        context.graph.should_receive("get_python_environment_marker").with_args(
            "tensorflow",
            "2.0.0",
            "https://pypi.org/simple",
            dependency_name="numpy",
            dependency_version="1.17.4",
            os_name="fedora",
            os_version="31",
            python_version="3.7",
        ).and_return("python_version >= '3.7'").once()
        product = Product.from_final_state(state=state, context=context)

        assert product.score == state.score
        assert product.justification == state.justification
        assert product.advised_runtime_environment == state.advised_runtime_environment
        assert product.project.to_dict() == {
            "requirements": {
                "packages": {
                    "daiquiri": {"index": "pypi-org", "version": "*"},
                    "tensorflow": {"index": "pypi-org", "version": ">=2.0.0"},
                },
                "dev-packages": {},
                "requires": {"python_version": "3.7"},
                "source": [{"url": "https://pypi.org/simple", "verify_ssl": True, "name": "pypi-org",}],
            },
            "requirements_locked": {
                "_meta": {
                    "sources": [{"url": "https://pypi.org/simple", "verify_ssl": True, "name": "pypi-org",}],
                    "requires": {"python_version": "3.7"},
                    "hash": {"sha256": "f08689732b596fd705a45bbf9ec44c3995b17a1aa6392c46500aeb736c4d4e88"},
                    "pipfile-spec": 6,
                },
                "default": {
                    "daiquiri": {"version": "==1.6.0", "hashes": ["sha256:000"], "index": "pypi-org",},
                    "numpy": {
                        "version": "==1.17.4",
                        "hashes": ["sha256:111"],
                        "index": "pypi-org",
                        "markers": "python_version >= '3.7'",
                    },
                    "tensorflow": {"version": "==2.0.0", "hashes": ["sha256:222"], "index": "pypi-org",},
                },
                "develop": {},
            },
            "runtime_environment": {
                "hardware": {"cpu_family": None, "cpu_model": None},
                "operating_system": {"name": "rhel", "version": None},
                "python_version": None,
                "cuda_version": None,
                "name": None,
                "platform": None,
            },
        }

    def test_to_dict(self) -> None:
        """Test conversion of this product into a dictionary representation."""
        project = flexmock()
        project.should_receive("to_dict").with_args().and_return({"baz": "bar"}).once()

        advised_runtime_environment = flexmock()
        advised_runtime_environment.should_receive("to_dict").with_args().and_return({"hello": "thoth"}).once()

        advised_manifest_changes = [
            [
                {
                    "apiVersion:": "apps.openshift.io/v1",
                    "kind": "DeploymentConfig",
                    "patch": {
                        "op": "add",
                        "path": "spec.template.spec.containers[0].env",
                        "value": {"name": "OMP_NUM_THREADS", "value": "1"},
                    },
                }
            ]
        ]

        product = Product(
            advised_manifest_changes=advised_manifest_changes,
            advised_runtime_environment=advised_runtime_environment,
            justification=[{"foo": "bar"}],
            project=project,
            score=0.999,
        )

        assert product.to_dict() == {
            "score": 0.999,
            "project": {"baz": "bar"},
            "justification": [{"foo": "bar"}],
            "advised_runtime_environment": {"hello": "thoth"},
            "advised_manifest_changes": advised_manifest_changes,
        }

    def test_environment_markers(self, context: Context) -> None:
        """Test handling of environment markers across multiple runs."""
        state = State(
            score=0.0,
            resolved_dependencies={
                "numpy": ("numpy", "1.0.0", "https://pypi.org/simple"),
                "tensorflow": ("tensorflow", "2.0.0", "https://pypi.org/simple"),
            },
            unresolved_dependencies={},
        )

        context.graph.should_receive("get_python_package_hashes_sha256").with_args(
            "numpy", "1.0.0", "https://pypi.org/simple"
        ).and_return(["000"]).once()

        context.graph.should_receive("get_python_package_hashes_sha256").with_args(
            "tensorflow", "2.0.0", "https://pypi.org/simple"
        ).and_return(["111"]).once()

        pypi = Source("https://pypi.org/simple")
        pv_numpy_locked = PackageVersion(name="numpy", version="==1.0.0", index=pypi, develop=False)
        pv_tensorflow_locked = PackageVersion(name="tensorflow", version="==2.0.0", index=pypi, develop=False)

        context.should_receive("get_package_version").with_args(
            ("numpy", "1.0.0", "https://pypi.org/simple"), graceful=False
        ).and_return(pv_numpy_locked).twice()

        context.should_receive("get_package_version").with_args(
            ("tensorflow", "2.0.0", "https://pypi.org/simple"), graceful=False
        ).and_return(pv_tensorflow_locked).twice()

        context.dependents = {
            "numpy": {
                ("numpy", "1.0.0", "https://pypi.org/simple"): {
                    (("tensorflow", "2.0.0", "https://pypi.org/simple"), "fedora", "31", "3.7",)
                }
            },
            "tensorflow": {("tensorflow", "2.0.0", "https://pypi.org/simple"): set()},
        }

        context.graph.should_receive("get_python_environment_marker").with_args(
            "tensorflow",
            "2.0.0",
            "https://pypi.org/simple",
            dependency_name="numpy",
            dependency_version="1.0.0",
            os_name="fedora",
            os_version="31",
            python_version="3.7",
        ).and_return("python_version >= '3.7'").and_return("python_version >= '3' or 1").twice()

        product = Product.from_final_state(context=context, state=state)
        expected = {
            "advised_manifest_changes": [],
            "advised_runtime_environment": None,
            "justification": [],
            "project": {
                "requirements": {
                    "dev-packages": {},
                    "packages": {"flask": "*", "tensorflow": "==1.9.0"},
                    "requires": {"python_version": "3.6"},
                    "source": [
                        {"name": "pypi", "url": "https://pypi.org/simple", "verify_ssl": True,},
                        {"name": "pypi-org", "url": "https://pypi.org/simple", "verify_ssl": True,},
                    ],
                },
                "requirements_locked": {
                    "_meta": {
                        "hash": {"sha256": "e55b6bbaba9467f1629c34e7a4180a6a2d82df37e02e762866e7aac27ced0f99"},
                        "pipfile-spec": 6,
                        "requires": {"python_version": "3.6"},
                        "sources": [
                            {"name": "pypi", "url": "https://pypi.org/simple", "verify_ssl": True,},
                            {"name": "pypi-org", "url": "https://pypi.org/simple", "verify_ssl": True,},
                        ],
                    },
                    "default": {
                        "numpy": {
                            "hashes": ["sha256:000"],
                            "index": "pypi-org",
                            "markers": "python_version >= '3.7'",
                            "version": "==1.0.0",
                        },
                        "tensorflow": {"hashes": ["sha256:111"], "index": "pypi-org", "version": "==2.0.0",},
                    },
                    "develop": {},
                },
                "runtime_environment": {
                    "cuda_version": None,
                    "hardware": {"cpu_family": None, "cpu_model": None},
                    "name": None,
                    "operating_system": {"name": None, "version": None},
                    "python_version": None,
                    "platform": None,
                },
            },
            "score": 0.0,
        }

        assert product.to_dict() == expected

        # Markers should not intersect.
        product = Product.from_final_state(context=context, state=state)
        expected["project"]["requirements_locked"]["default"]["numpy"]["markers"] = "python_version >= '3' or 1"
        assert product.to_dict() == expected



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test implementation of a predictor used to generate package combinations faster."""

import pytest

from thoth.adviser.predictors import PackageCombinations

from ..base import AdviserTestCase


class TestPackageCombinations(AdviserTestCase):
    """Test implementation of a predictor used to generate package combinations faster."""

    def test_pre_run_error(self) -> None:
        """Test pre-run initialization failure."""
        predictor = PackageCombinations()
        with pytest.raises(ValueError, match="No package combinations supplied to the predictor"):
            predictor.pre_run()

        predictor = PackageCombinations(package_combinations=[])
        with pytest.raises(ValueError, match="No package combinations supplied to the predictor"):
            predictor.pre_run()

    def test_pre_run(self) -> None:
        """Test pre-run initialization."""
        predictor = PackageCombinations(package_combinations=["tensorflow", "numpy"])
        assert isinstance(predictor.package_combinations, set)
        assert predictor.package_combinations == {"tensorflow", "numpy"}



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test Dependency Monkey implementation."""

from typing import Any
from typing import Dict
from typing import List
from typing import Optional
import random
import sys
import json

import flexmock
import amun

from thoth.adviser.enums import DecisionType
from thoth.adviser.product import Product
from thoth.adviser.pipeline_config import PipelineConfig
from thoth.adviser.resolver import Resolver
from thoth.adviser.dependency_monkey import DependencyMonkey
from thoth.adviser.dm_report import DependencyMonkeyReport

from .base import AdviserTestCase


class TestDependencyMonkey(AdviserTestCase):
    """Tests related to Dependency Monkey."""

    @staticmethod
    def _get_test_dm(
        *, stack_output: str, with_devel: bool, products: List[Product], amun_context: Optional[Dict[str, Any]] = None
    ) -> DependencyMonkey:
        """Get instantiated dependency monkey ready to be tested."""
        flexmock(Resolver)
        (Resolver.should_receive("resolve_products").with_args(with_devel=with_devel).and_return(products).once())

        flexmock(PipelineConfig)
        (PipelineConfig.should_receive("call_post_run_report").and_return(None).once())

        dependency_monkey = DependencyMonkey(
            resolver=Resolver(pipeline=PipelineConfig(), project=None, library_usage=None, graph=None, predictor=None,),
            stack_output=stack_output,
            decision_type=DecisionType.ALL,
            context=amun_context or {},
        )

        return dependency_monkey

    @staticmethod
    def _get_product_mock() -> Any:
        """Get a mocked product for testing."""

    def test_construct(self) -> None:
        """Test obtaining predictor from Dependency Monkey instance."""
        resolver = flexmock(predictor="foo")
        dependency_monkey = DependencyMonkey(
            resolver=resolver,
            stack_output="-",
            context={"bar": "baz"},
            dry_run=True,
            decision_type=DecisionType.RANDOM,
        )

        assert dependency_monkey.stack_output == "-"
        assert dependency_monkey.context == {"bar": "baz"}
        assert dependency_monkey.dry_run is True
        assert dependency_monkey.decision_type is DecisionType.RANDOM

        # This test actually just makes sure we have predictor property available :)
        assert dependency_monkey.predictor is resolver.predictor

    def test_amun_output(self) -> None:
        """Test Amun API stack submissions."""
        project = flexmock()
        generated_project_dict = {"bar": 1}
        project.should_receive("to_dict").with_args().and_return(generated_project_dict).once()

        product = flexmock(
            project=project,
            score=random.random(),
            justification=[{"justification": "some another justification"}],
            advised_runtime_environment=flexmock(),
        )
        product_dict = {"baz": 2}
        product.should_receive("to_dict").with_args().and_return(product_dict).once()

        amun_api = "http://amun-api"
        amun_context = {"base": "ubi:8"}
        flexmock(amun).should_receive("inspect").with_args(
            amun_api, base=amun_context["base"], python=generated_project_dict
        ).and_return({"inspection_id": "inspection-deadbeef"}).once()

        dependency_monkey = self._get_test_dm(
            stack_output=amun_api, with_devel=True, products=[product], amun_context=amun_context,
        )

        report: DependencyMonkeyReport = dependency_monkey.resolve(with_devel=True)
        assert report.to_dict() == {
            "skipped": 0,
            "responses": [{"response": "inspection-deadbeef", "product": product_dict}],
        }

    def test_dir_output(self) -> None:
        """Test writing output to a directory."""
        project = flexmock()
        project.should_receive("to_files").with_args("/tmp/1/Pipfile", "/tmp/1/Pipfile.lock").and_return(
            {"foo": "bar"}
        ).once()

        product = flexmock(
            project=project,
            score=random.random(),
            justification=[{"justification": "some justification"}],
            advised_runtime_environment=flexmock(),
        )
        product_dict = {"baz": 2}
        product.should_receive("to_dict").with_args().and_return(product_dict).once()

        dependency_monkey = self._get_test_dm(stack_output="/tmp", with_devel=False, products=[product])
        report: DependencyMonkeyReport = dependency_monkey.resolve(with_devel=False)

        assert report.to_dict() == {
            "skipped": 0,
            "responses": [{"response": "/tmp/1", "product": product_dict}],
        }

    def test_stdout_output(self) -> None:
        """Test writing output to standard output."""
        project = flexmock()
        generated_project_dict = {"bar": 1}
        project.should_receive("to_dict").with_args().and_return(generated_project_dict).once()

        product = flexmock(
            project=project,
            score=random.random(),
            justification=[{"justification": "some justification"}],
            advised_runtime_environment=flexmock(),
        )

        dependency_monkey = self._get_test_dm(stack_output="-", with_devel=True, products=[product])
        flexmock(json).should_receive("dump").with_args(
            generated_project_dict, fp=sys.stdout, sort_keys=True, indent=2
        ).and_return(None).once()
        report: DependencyMonkeyReport = dependency_monkey.resolve(with_devel=True)

        # When stdout is used, products are not carried in the final report.
        assert report.to_dict() == {"skipped": 0, "responses": []}



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A base class for implementing adviser's test cases."""

from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
import os

from voluptuous import All
from voluptuous import Any as SchemaAny
from voluptuous import Optional as SchemaOptional
from voluptuous import Invalid
from voluptuous import Length
from voluptuous import Schema


class AdviserTestCaseException(Exception):
    """A base class for exceptions that can occur in the test suite."""


class AdviserJustificationSchemaError(AdviserTestCaseException):
    """An exception raised when the justification reported violates schema."""


class AdviserTestCase:
    """A base class for implementing adviser's test cases."""

    data_dir = Path(os.path.dirname(os.path.realpath(__file__))) / "data"

    JUSTIFICATION_SAMPLE_1 = [
        {"message": "Justification sample 1", "type": "WARNING", "link": "https://thoth-station.ninja"},
        {"message": "Justification sample 1", "type": "INFO", "link": "https://thoth-station.ninja"},
        {"message": "Justification sample 1", "type": "ERROR", "link": "https://thoth-station.ninja"},
    ]

    JUSTIFICATION_SAMPLE_2 = [
        {
            "message": "Justification sample 2",
            "type": "INFO",
            "link": "https://thoth-station.ninja",
            "advisory": "Bark!",
        },
    ]

    JUSTIFICATION_SAMPLE_3 = [
        {
            "message": "Justification sample 2",
            "type": "INFO",
            "link": "https://thoth-station.ninja",
            "package_name": "tensorflow",
            "version_range": "<2.3>=",
        },
    ]

    _JUSTIFICATION_SCHEMA = Schema(
        [
            {
                "message": All(str, Length(min=1)),
                "type": SchemaAny("INFO", "WARNING", "ERROR", "LATEST", "CVE"),
                "link": All(str, Length(min=1)),
                SchemaOptional("advisory"): All(str, Length(min=1)),
                SchemaOptional("cve_id"): All(str, Length(min=1)),
                SchemaOptional("cve_name"): All(str, Length(min=1)),
                SchemaOptional("package_name"): All(str, Length(min=1)),
                SchemaOptional("version_range"): All(str, Length(min=1)),
            }
        ]
    )

    @classmethod
    def verify_justification_schema(cls, justification: Optional[List[Dict[str, Any]]]) -> bool:
        """Verify the justification schema is correct."""
        if justification is None:
            return True

        try:
            cls._JUSTIFICATION_SCHEMA(justification)
        except Invalid as exc:
            raise AdviserJustificationSchemaError(exc.msg) from exc
        else:
            return True



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test removing pinned versions in direct dependencies."""

import flexmock

from thoth.adviser.sieves import CutLockedSieve
from thoth.python import Source
from thoth.python import PackageVersion
from thoth.python import Project

from ..base import AdviserTestCase


class TestCutPreReleasesSieve(AdviserTestCase):
    """Test removing pinned versions in direct dependencies."""

    _CASE_PIPFILE_LOCKED = """
[[source]]
url = "https://pypi.org/simple"
verify_ssl = true
name = "pypi"

[packages]
tensorflow = "==1.9.0"

[dev-packages]
pytest = "==5.3.1"
"""

    _CASE_PIPFILE_NOT_LOCKED = """
[[source]]
url = "https://pypi.org/simple"
verify_ssl = true
name = "pypi"

[packages]
tensorflow = "*"

[dev-packages]
pytest = "*"

[pipenv]
allow_prereleases = true
"""

    def test_cut_locked(self) -> None:
        """Test removing a locked package based on direct dependencies."""
        tf = PackageVersion(
            name="tensorflow", version="==2.0.0", index=Source("https://pypi.org/simple"), develop=False,
        )

        context = flexmock(project=Project.from_strings(self._CASE_PIPFILE_LOCKED))
        with CutLockedSieve.assigned_context(context):
            sieve = CutLockedSieve()
            assert list(sieve.run(p for p in [tf])) == []

    def test_cut_locked_dev(self) -> None:
        """Test removing a locked package based on direct dev dependencies."""
        pytest = PackageVersion(
            name="pytest", version="==2.0.0", index=Source("https://pypi.org/simple"), develop=False,
        )

        context = flexmock(project=Project.from_strings(self._CASE_PIPFILE_LOCKED))
        with CutLockedSieve.assigned_context(context):
            sieve = CutLockedSieve()
            assert list(sieve.run(p for p in [pytest])) == []

    def test_no_cut(self) -> None:
        """Test not removing a locked package based on direct dependencies."""
        tf = PackageVersion(
            name="tensorflow", version="==1.9.0", index=Source("https://pypi.org/simple"), develop=False,
        )

        context = flexmock(project=Project.from_strings(self._CASE_PIPFILE_LOCKED))
        with CutLockedSieve.assigned_context(context):
            sieve = CutLockedSieve()
            assert list(sieve.run(p for p in [tf])) == [tf]

    def test_no_cut_dev(self) -> None:
        """Test not removing a locked package based on dev direct dependencies."""
        pytest = PackageVersion(
            name="pytest", version="==5.3.1", index=Source("https://pypi.org/simple"), develop=False,
        )

        context = flexmock(project=Project.from_strings(self._CASE_PIPFILE_LOCKED))
        with CutLockedSieve.assigned_context(context):
            sieve = CutLockedSieve()
            assert list(sieve.run(p for p in [pytest])) == [pytest]

    def test_noop(self) -> None:
        """Test no operation if dependencies are not locked."""
        tf = PackageVersion(
            name="tensorflow", version="==1.9.0", index=Source("https://pypi.org/simple"), develop=False,
        )

        context = flexmock(project=Project.from_strings(self._CASE_PIPFILE_NOT_LOCKED))
        with CutLockedSieve.assigned_context(context):
            sieve = CutLockedSieve()
            assert list(sieve.run(p for p in [tf])) == [tf]

    def test_noop_dev(self) -> None:
        """Test no operation if dependencies are not locked."""
        pytest = PackageVersion(
            name="pytest", version="==5.3.1", index=Source("https://pypi.org/simple"), develop=False,
        )

        context = flexmock(project=Project.from_strings(self._CASE_PIPFILE_NOT_LOCKED))
        with CutLockedSieve.assigned_context(context):
            sieve = CutLockedSieve()
            assert list(sieve.run(p for p in [pytest])) == [pytest]



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A boot to check for solved software environment before running any resolution."""

import logging
from typing import Optional
from typing import Dict
from typing import Any
from typing import TYPE_CHECKING

import attr
from thoth.common import get_justification_link as jl

from ..boot import Boot
from ..exceptions import NotAcceptable

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class SolvedSoftwareEnvironmentBoot(Boot):
    """A boot to check for solved software environment before running any resolution."""

    _JUSTIFICATION_LINK = jl("solved_sw_env")

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Optional[Dict[str, Any]]:
        """Register self, always."""
        if not builder_context.project.runtime_environment.is_fully_specified():
            return None

        if not builder_context.is_included(cls):
            return {}

        return None

    def run(self) -> None:
        """Check for version clash in packages."""
        if self.context.graph.solved_software_environment_exists(
            os_name=self.context.project.runtime_environment.operating_system.name,
            os_version=self.context.project.runtime_environment.operating_system.version,
            python_version=self.context.project.runtime_environment.python_version,
        ):
            return

        runtime_environment = self.context.project.runtime_environment
        msg = (
            f"No observations found for {runtime_environment.operating_system.name!r} in "
            f"version {runtime_environment.operating_system.version!r} using "
            f"Python {runtime_environment.python_version!r} - see {self._JUSTIFICATION_LINK}"
        )

        _LOGGER.warning("%s - %s", msg, self._JUSTIFICATION_LINK)
        _LOGGER.warning("Available configurations:")

        configurations = self.context.graph.get_solved_python_package_versions_software_environment_all()
        _LOGGER.warning("{:<16} {:<16} {:<8}".format("OS name", "OS version", "Python version"))
        for conf in sorted(configurations, key=lambda i: (i["os_name"], i["os_version"], i["python_version"]),):
            _LOGGER.warning("{:<16} {:<16} {:<8}".format(conf["os_name"], conf["os_version"], conf["python_version"]))

        raise NotAcceptable(msg)



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Kevin Postlethwait
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test filtering out Python Packages based on required and available ABI symbols."""

import flexmock

from thoth.adviser.sieves import AbiCompatibilitySieve
from thoth.python import PackageVersion
from thoth.python import Source
from thoth.storages import GraphDatabase

from ..base import AdviserTestCase

_SYSTEM_SYMBOLS = ["GLIBC_2.0", "GLIBC_2.1", "GLIBC_2.2", "GLIBC_2.3", "GLIBC_2.4", "GLIBC_2.5", "GCC_3.4", "X_2.21"]
_REQUIRED_SYMBOLS_A = ["GLIBC_2.9"]
_REQUIRED_SYMBOLS_B = ["GLIBC_2.4"]


class TestAbiCompatSieve(AdviserTestCase):
    """Test filtering out packages based on symbols required."""

    def test_abi_compat_symbols_present(self) -> None:
        """Test if required symbols are correctly identified."""
        source = Source("https://pypi.org/simple")
        package_version = PackageVersion(name="tensorflow", version="==1.9.0", index=source, develop=False)
        flexmock(GraphDatabase)
        GraphDatabase.should_receive("get_analyzed_image_symbols_all").and_return(_SYSTEM_SYMBOLS).once()
        GraphDatabase.should_receive("get_python_package_required_symbols").and_return(_REQUIRED_SYMBOLS_B).once()

        context = flexmock(
            graph=GraphDatabase(),
            project=flexmock(
                runtime_environment=flexmock(
                    operating_system=flexmock(name="rhel", version="8.0"), cuda_version="4.6", python_version="3.6",
                )
            ),
        )
        with AbiCompatibilitySieve.assigned_context(context):
            sieve = AbiCompatibilitySieve()
            sieve.pre_run()
            assert list(sieve.run((p for p in [package_version]))) == [package_version]

    def test_abi_compat_symbols_not_present(self) -> None:
        """Test if required symbols being missing is correctly identified."""
        source = Source("https://pypi.org/simple")
        package_version = PackageVersion(name="tensorflow", version="==1.9.0", index=source, develop=False)
        flexmock(GraphDatabase)
        GraphDatabase.should_receive("get_analyzed_image_symbols_all").and_return(_SYSTEM_SYMBOLS).once()
        GraphDatabase.should_receive("get_python_package_required_symbols").and_return(_REQUIRED_SYMBOLS_A).once()

        context = flexmock(
            graph=GraphDatabase,
            project=flexmock(
                runtime_environment=flexmock(
                    operating_system=flexmock(name="rhel", version="8.0"), cuda_version="4.6", python_version="3.6",
                )
            ),
        )
        with AbiCompatibilitySieve.assigned_context(context):
            sieve = AbiCompatibilitySieve()
            sieve.pre_run()
            assert list(sieve.run((p for p in [package_version]))) == []



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test boot for Python version assignment."""

import flexmock

from thoth.adviser.boots import PythonVersionBoot
from thoth.common import get_justification_link as jl
from thoth.python import Project

from ..base import AdviserTestCase


class TestPythonVersionBoot(AdviserTestCase):
    """Test changes to runtime environment or configuration done with respect to configured Python."""

    _CASE_PIPFILE_NO_PYTHON = """
[[source]]
url = "https://pypi.org/simple"
verify_ssl = true
name = "pypi"

[packages]
tensorflow = "*"
"""

    _CASE_PIPFILE_PYTHON = """
[[source]]
url = "https://pypi.org/simple"
verify_ssl = true
name = "pypi"

[packages]
tensorflow = "*"

[requires]
python_version = "3.6"
"""

    def test_no_python_config(self) -> None:
        """Test assigning Python version from Thoth's config file."""
        context = flexmock(project=Project.from_strings(self._CASE_PIPFILE_PYTHON), stack_info=[])
        context.project.runtime_environment.operating_system.name = "rhel"
        context.project.runtime_environment.operating_system.version = "8"

        assert context.project.runtime_environment.python_version is None

        boot = PythonVersionBoot()
        with PythonVersionBoot.assigned_context(context):
            boot.run()

        assert context.project.runtime_environment.operating_system.name == "rhel"
        assert context.project.runtime_environment.operating_system.version == "8"
        assert context.project.runtime_environment.python_version == "3.6"
        assert context.stack_info == [
            {
                "message": "No version of Python specified in the configuration, using "
                "Python version found in Pipfile: '3.6'",
                "type": "WARNING",
                "link": jl("py_version"),
            },
        ]

    def test_no_python_pipfile(self) -> None:
        """Test assigning Python version from Pipfile."""
        context = flexmock(project=Project.from_strings(self._CASE_PIPFILE_NO_PYTHON), stack_info=[])
        context.project.runtime_environment.operating_system.name = "rhel"
        context.project.runtime_environment.operating_system.version = "8"
        context.project.runtime_environment.python_version = "3.6"

        boot = PythonVersionBoot()
        with PythonVersionBoot.assigned_context(context):
            boot.run()

        assert context.project.runtime_environment.operating_system.name == "rhel"
        assert context.project.runtime_environment.operating_system.version == "8"
        assert context.project.runtime_environment.python_version == "3.6"
        assert context.stack_info == [
            {
                "message": "No version of Python specified explicitly, assigning the one "
                "found in Thoth's configuration: '3.6'",
                "type": "WARNING",
                "link": jl("py_version"),
            }
        ]



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2018, 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Fetcher for fetching digests from the graph database."""

import logging
from typing import List
from typing import Dict

import attr
from thoth.storages import GraphDatabase
from thoth.python import DigestsFetcherBase


_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class GraphDigestsFetcher(DigestsFetcherBase):  # type: ignore
    """Fetch digests from the graph database."""

    graph = attr.ib(type=GraphDatabase)

    @graph.default
    def _graph_default(self) -> GraphDatabase:
        """Get default graph instance if no explicitly provided."""
        graph = GraphDatabase()
        graph.connect()
        return graph

    def fetch_digests(self, package_name: str, package_version: str) -> Dict[str, List[Dict[str, str]]]:
        """Fetch digests for the given package in specified version, consider only enabled indexes."""
        _LOGGER.debug(
            "Querying graph database for digests for package %r in version %r", package_name, package_version,
        )

        result = {}
        for index_url in self.graph.get_python_package_index_urls_all(enabled=True):
            query_result = self.graph.get_python_package_hashes_sha256(
                package_name, package_version, index_url, distinct=True
            )
            result[index_url] = [{"sha256": digest} for digest in query_result]

        return result



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Recommend specific TensorFlow release based on CUDA version present in the runtime environment."""

import attr
from typing import Any
from typing import Optional
from typing import Tuple
from typing import Generator
from typing import Dict
from typing import Set
from typing import FrozenSet
from typing import TYPE_CHECKING
import logging

from thoth.common import get_justification_link as jl
from thoth.python import PackageVersion

from ...enums import RecommendationType
from ...sieve import Sieve


if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext


_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class TensorFlowCUDASieve(Sieve):
    """A sieve that makes sure the right TensorFlow version is used when CUDA is present in the runtime environment.

    See supported matrix at https://www.tensorflow.org/install/source#linux
    """

    _MESSAGE = f"Recommended TensorFlow that supports CUDA present in the runtime environment - see {jl('tf_cuda')}"

    _EMPTY: FrozenSet[Tuple[int, int]] = frozenset()
    _TF_1_CUDA_8_SUPPORT = frozenset({(1, 0), (1, 1), (1, 2), (1, 3), (1, 4)})
    _TF_1_CUDA_9_SUPPORT = frozenset({(1, 5), (1, 6), (1, 7), (1, 8), (1, 9), (1, 10), (1, 11), (1, 12)})
    _TF_1_CUDA_10_0_SUPPORT = frozenset({(1, 13), (1, 14), (1, 15)})
    _TF_2_CUDA_10_0_SUPPORT = frozenset({(2, 0)})
    _TF_2_CUDA_10_1_SUPPORT = frozenset({(2, 1), (2, 2), (2, 3)})
    _KNOWN_CUDA = frozenset({"8", "9", "10.0", "10.1"})

    # Holds tensorflow version for which a message was printed to logs.
    _messages_logged = attr.ib(type=Set[str], default=attr.Factory(set), init=False)
    # Holds a frozen set with versions of TensorFlow supported.
    _tf_1_cuda_support = attr.ib(type=FrozenSet[Tuple[int, int]], default=_EMPTY, init=False)
    _tf_2_cuda_support = attr.ib(type=FrozenSet[Tuple[int, int]], default=_EMPTY, init=False)

    def pre_run(self) -> None:
        """Initialize this pipeline unit before each run."""
        self._messages_logged.clear()
        cuda_version = self.context.project.runtime_environment.cuda_version

        if cuda_version == "8":
            self._tf_1_cuda_support = self._TF_1_CUDA_8_SUPPORT
            self._tf_2_cuda_support = self._EMPTY
        elif cuda_version == "9":
            self._tf_1_cuda_support = self._TF_1_CUDA_9_SUPPORT
            self._tf_2_cuda_support = self._EMPTY
        elif cuda_version == "10.0":
            self._tf_1_cuda_support = self._TF_1_CUDA_10_0_SUPPORT
            self._tf_2_cuda_support = self._TF_2_CUDA_10_0_SUPPORT
        elif cuda_version == "10.1":
            self._tf_1_cuda_support = self._EMPTY
            self._tf_2_cuda_support = self._TF_2_CUDA_10_1_SUPPORT
        else:
            _LOGGER.error("Unsupported CUDA version, cannot provide recommendations for TensorFlow")
            self._tf_1_cuda_support = self._EMPTY
            self._tf_2_cuda_support = self._EMPTY

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Optional[Dict[str, Any]]:
        """Register this pipeline unit for adviser when CUDA is present."""
        if not builder_context.is_adviser_pipeline():
            return None

        if builder_context.recommendation_type in (RecommendationType.LATEST, RecommendationType.TESTING):
            # Use any TensorFlow for testing purposes or when resolving latest stack.
            return None

        cuda_version = builder_context.project.runtime_environment.cuda_version
        if cuda_version is None:
            # No CUDA available in the given runtime environment.
            return None

        if cuda_version not in cls._KNOWN_CUDA:
            _LOGGER.warning(
                "Unable to perform recommendations for TensorFlow based on CUDA version: "
                "unknown CUDA version supplied - see %s",
                jl("tf_unknown_cuda"),
            )
            return None

        if not builder_context.is_included(cls):
            return {}

        return None

    def _log_unknown_tf_version(self, package_version: PackageVersion) -> None:
        """Log an unhandled TensorFlow release, this pipeline unit needs an update in such cases."""
        message = (
            f"Unhandled TensorFlow release {package_version.to_tuple()}, gracefully giving up recommending "
            f"TensorFlow based on CUDA version {self.context.project.runtime_environment.cuda_version!r}"
        )
        _LOGGER.error("%s - see %s", message, jl("cuda_unknown_tf"))
        self.context.stack_info.append(
            {"type": "ERROR", "message": message, "link": jl("cuda_unknown_tf"),}
        )

    def _maybe_log_no_recommendations(self, package_version: PackageVersion) -> None:
        """Log no recommendations are given for the given configuration."""
        if package_version.locked_version not in self._messages_logged:
            message = (
                f"Not recommending TensorFlow in version {package_version.locked_version!r} as this version "
                f"is not supported on CUDA in version {self.context.project.runtime_environment.cuda_version!r}"
            )
            link = jl("tf_no_cuda")
            _LOGGER.warning("%s - see %s", message, link)
            self._messages_logged.add(package_version.locked_version)
            self.context.stack_info.append({"type": "ERROR", "message": message, "link": link})

    def _yield_tensorflow(self, package_version: PackageVersion) -> bool:
        """Handle a tensorflow release."""
        tf_semantic_version = package_version.semantic_version.release[:2]

        if tf_semantic_version[0] == 1:
            self._maybe_log_no_recommendations(package_version)
            return False
        elif tf_semantic_version <= (2, 3):
            if tf_semantic_version in self._tf_2_cuda_support:
                return True

            self._maybe_log_no_recommendations(package_version)
            return False

        self._log_unknown_tf_version(package_version)
        return True

    def _yield_tensorflow_gpu(self, package_version: PackageVersion) -> bool:
        """Handle a tensorflow-gpu release."""
        if package_version.semantic_version.release[0] == 1:
            if package_version.semantic_version.release[:2] in self._tf_1_cuda_support:
                return True

            self._maybe_log_no_recommendations(package_version)
            return False
        elif package_version.semantic_version.release[0] == 2:
            # tensorflow-gpu>=2 acts as tensorflow>=2
            return self._yield_tensorflow(package_version)

        self._log_unknown_tf_version(package_version)
        return True

    def run(self, package_versions: Generator[PackageVersion, None, None]) -> Generator[PackageVersion, None, None]:
        """Use specific TensorFlow release based on CUDA version present in the runtime environment."""
        for package_version in package_versions:
            if package_version.name == "tensorflow":
                if self._yield_tensorflow(package_version):
                    yield package_version
            elif package_version.name == "tensorflow-gpu":
                if self._yield_tensorflow_gpu(package_version):
                    yield package_version



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test suggesting not to use TensorFlow 2.2 with tensorflow-probability."""

import pytest

from thoth.adviser.context import Context
from thoth.adviser.enums import DecisionType
from thoth.adviser.enums import RecommendationType
from thoth.adviser.exceptions import NotAcceptable
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.state import State
from thoth.adviser.steps import TensorFlow22ProbabilityStep
from thoth.python import PackageVersion
from thoth.python import Source

from ...base import AdviserTestCase


class TestTensorFlow22ProbabilityStep(AdviserTestCase):
    """Test suggesting not to use TensorFlow 2.2 with tensorflow-probability."""

    @pytest.mark.parametrize("recommendation_type", [RecommendationType.STABLE, RecommendationType.TESTING])
    def test_include(self, builder_context: PipelineBuilderContext, recommendation_type: RecommendationType) -> None:
        """Test including this pipeline unit."""
        builder_context.decision_type = None
        builder_context.recommendation_type = recommendation_type
        assert builder_context.is_adviser_pipeline()
        assert TensorFlow22ProbabilityStep.should_include(builder_context) == {}

    @pytest.mark.parametrize(
        "recommendation_type,decision_type",
        [(RecommendationType.LATEST, None), (None, DecisionType.RANDOM),],  # A Dependency Monkey run.
    )
    def test_no_include(
        self, builder_context: PipelineBuilderContext, recommendation_type, decision_type: DecisionType,
    ) -> None:
        """Test not including this pipeline unit step."""
        builder_context.decision_type = decision_type
        builder_context.recommendation_type = recommendation_type
        assert TensorFlow22ProbabilityStep.should_include(builder_context) is None

    @pytest.mark.parametrize("package_name", ["tensorflow", "tensorflow-cpu", "tensorflow-gpu"])
    def test_run(self, context: Context, package_name: str) -> None:
        """Test recommending not to use TensorFlow 2.2 with tensorflow-probability."""
        package_version = PackageVersion(
            name="tensorflow-probability", version="==0.11.0", develop=False, index=Source("https://pypi.org/simple"),
        )

        state = State()
        state.add_resolved_dependency((package_name, "2.2.0", "https://pypi.org/simple"))

        unit = TensorFlow22ProbabilityStep()
        unit.pre_run()

        assert not context.stack_info

        with unit.assigned_context(context):
            with pytest.raises(NotAcceptable):
                assert unit._message_logged is False
                unit.run(state, package_version)
                assert unit._message_logged is True

        assert context.stack_info
        assert self.verify_justification_schema(context.stack_info) is True

    def test_run_acceptable_tf(self) -> None:
        """Test noop for this pipeline unit."""
        package_version_1 = PackageVersion(
            name="tensorflow-probability", version="==0.11.0", develop=False, index=Source("https://pypi.org/simple"),
        )

        package_version_2 = PackageVersion(
            name="flask", version="==0.12", develop=False, index=Source("https://pypi.org/simple"),
        )

        state = State()
        unit = TensorFlow22ProbabilityStep()

        assert unit.run(state, package_version_1) is None
        assert unit.run(state, package_version_2) is None

        state.add_resolved_dependency(("tensorflow", "2.3.0", "https://pypi.org/simple"))
        assert unit.run(state, package_version_1) is None
        assert unit.run(state, package_version_2) is None



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test sieve to filter out old setuptools that do not work with Python 3.6."""

import flexmock
import pytest

from thoth.adviser.enums import DecisionType
from thoth.adviser.enums import RecommendationType
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.sieves import Py36SetuptoolsSieve
from thoth.adviser.context import Context
from thoth.python import PackageVersion
from thoth.python import Source

from ...base import AdviserTestCase


class TestPy36SetuptoolsSieve(AdviserTestCase):
    """Test sieve to filter out old setuptools that do not work with Python 3.6."""

    @pytest.mark.parametrize(
        "recommendation_type,decision_type",
        [
            (RecommendationType.STABLE, None),
            (RecommendationType.TESTING, None),
            (None, DecisionType.RANDOM),
            (None, DecisionType.ALL),
        ],
    )
    def test_include(
        self,
        builder_context: PipelineBuilderContext,
        recommendation_type: RecommendationType,
        decision_type: DecisionType,
    ) -> None:
        """Test including this pipeline unit."""
        builder_context.decision_type = decision_type
        builder_context.recommendation_type = recommendation_type
        builder_context.project.runtime_environment.python_version = "3.6"

        assert builder_context.is_adviser_pipeline() or builder_context.is_dependency_monkey_pipeline()
        assert Py36SetuptoolsSieve.should_include(builder_context) == {}

    @pytest.mark.parametrize(
        "recommendation_type,decision_type,python_version",
        [
            (RecommendationType.STABLE, None, "3.8"),
            (RecommendationType.TESTING, None, "3.9"),
            (None, DecisionType.RANDOM, "3.5"),
        ],
    )
    def test_no_include(
        self,
        builder_context: PipelineBuilderContext,
        recommendation_type: RecommendationType,
        decision_type: DecisionType,
        python_version: str,
    ) -> None:
        """Test not including this pipeline unit step."""
        builder_context.decision_type = decision_type
        builder_context.recommendation_type = recommendation_type
        builder_context.project.runtime_environment.python_version = python_version
        assert Py36SetuptoolsSieve.should_include(builder_context) is None

    def test_filter(self, context: Context) -> None:
        """Test filtering out setuptools that do not work with Python 3.6."""
        package_version = PackageVersion(
            name="setuptools", version="==14.2", develop=False, index=Source("https://pypi.org/simple"),
        )

        unit = Py36SetuptoolsSieve()
        unit.pre_run()
        with Py36SetuptoolsSieve.assigned_context(context):
            assert unit._message_logged is False
            assert list(unit.run(p for p in [package_version])) == []
            assert unit._message_logged is True
            assert context.stack_info, "No stack guidance provided"
            assert self.verify_justification_schema(context.stack_info)

    def test_no_filter(self) -> None:
        """Test not filtering packages that can be included."""
        pkg1 = PackageVersion(
            name="setuptools", version="==17.0", develop=False, index=Source("https://pypi.org/simple"),
        )
        pkg2 = PackageVersion(
            name="setuptools", version="==49.1.0", develop=False, index=Source("https://pypi.org/simple"),
        )
        pkg3 = PackageVersion(
            name="tensorflow", version="==2.2.0", develop=False, index=Source("https://thoth-station.ninja/simple"),
        )

        pkgs = [pkg1, pkg2, pkg3]

        context = flexmock()
        with Py36SetuptoolsSieve.assigned_context(context):
            unit = Py36SetuptoolsSieve()
            assert list(unit.run(p for p in pkgs)) == pkgs



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A step to filter out packages that are pinned to a specific version."""

import logging
from typing import Optional
from typing import Dict
from typing import Any
from typing import Generator
from typing import TYPE_CHECKING

import attr
from thoth.python import PackageVersion

from ..sieve import Sieve

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class CutLockedSieve(Sieve):
    """Cut-off packages that are locked to a specific version.

    If a project pins down a package to a specific release, respect that. Otherwise
    resolver does not need to find any resolved stack, especially considering only
    N latest versions and the pinned version is >=N+1 version.
    """

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Optional[Dict[str, Any]]:
        """Include cut-locked pipeline sieve for adviser or Dependency Monkey, always."""
        if not builder_context.is_included(cls):
            return {}

        return None

    def run(self, package_versions: Generator[PackageVersion, None, None]) -> Generator[PackageVersion, None, None]:
        """Cut-off locked versions to a specific version."""
        packages = self.context.project.pipfile.packages.packages
        dev_packages = self.context.project.pipfile.dev_packages.packages

        for package_version in package_versions:
            direct_package = packages.get(package_version.name)
            direct_dev_package = dev_packages.get(package_version.name)

            if direct_package is None and direct_dev_package is None:
                yield package_version

            if (
                direct_package
                and direct_package.is_locked()
                and direct_package.locked_version != package_version.locked_version
            ) or (
                direct_dev_package
                and direct_dev_package.is_locked()
                and direct_dev_package.locked_version != package_version.locked_version
            ):
                _LOGGER.debug(
                    "Removing package %s - it does not correspond to package version locked by direct dependencies",
                    package_version.to_tuple(),
                )
                continue

            yield package_version



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test a step for TensorFlow <2.3>=2.0 that states NumPy<2.0.0 as a dependency, but is compatible with NumPy<1.19.0."""

from itertools import product

import pytest

from thoth.adviser.enums import DecisionType
from thoth.adviser.enums import RecommendationType
from thoth.adviser.exceptions import NotAcceptable
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.state import State
from thoth.adviser.context import Context
from thoth.adviser.steps import TensorFlow22NumPyStep
from thoth.python import PackageVersion
from thoth.python import Source

from ...base import AdviserTestCase


class TestTensorFlow22NumPyStep(AdviserTestCase):
    """TensorFlow <2.3>=2.0 states NumPy<2.0.0 as a dependency, but is compatible with NumPy<1.19.0."""

    @pytest.mark.parametrize(
        "recommendation_type",
        [
            RecommendationType.STABLE,
            RecommendationType.TESTING,
            RecommendationType.PERFORMANCE,
            RecommendationType.STABLE,
        ],
    )
    def test_include(self, builder_context: PipelineBuilderContext, recommendation_type: RecommendationType) -> None:
        """Test including this pipeline unit."""
        builder_context.decision_type = None
        builder_context.recommendation_type = recommendation_type
        assert builder_context.is_adviser_pipeline()
        assert TensorFlow22NumPyStep.should_include(builder_context) == {}

    @pytest.mark.parametrize(
        "recommendation_type,decision_type",
        [(RecommendationType.LATEST, None), (None, DecisionType.RANDOM),],  # A Dependency Monkey run.
    )
    def test_no_include(
        self, builder_context: PipelineBuilderContext, recommendation_type, decision_type: DecisionType,
    ) -> None:
        """Test not including this pipeline unit step."""
        builder_context.decision_type = decision_type
        builder_context.recommendation_type = recommendation_type
        assert TensorFlow22NumPyStep.should_include(builder_context) is None

    @pytest.mark.parametrize(
        "tf_name,tf_version,np_version",
        list(
            product(
                ["tensorflow", "tensorflow-cpu", "tensorflow-gpu", "intel-tensorflow"],
                ["2.0.0", "2.1.0", "2.2.0"],
                ["1.19.0", "1.20.0"],
            )
        ),
    )
    def test_run_not_acceptable(self, context: Context, tf_name: str, tf_version: str, np_version: str) -> None:
        """Test resolutions that are not acceptable."""
        package_version = PackageVersion(
            name="numpy", version=f"=={np_version}", develop=False, index=Source("https://pypi.org/simple"),
        )

        tf_package_version = PackageVersion(
            name=tf_name, version=f"=={tf_version}", develop=False, index=Source("https://pypi.org/simple"),
        )

        state = State()
        state.add_resolved_dependency(tf_package_version.to_tuple())
        context.register_package_version(tf_package_version)

        unit = TensorFlow22NumPyStep()
        unit.pre_run()

        with unit.assigned_context(context):
            assert unit._message_logged is False
            with pytest.raises(NotAcceptable):
                unit.run(state, package_version)

        assert unit._message_logged is True

    @pytest.mark.parametrize(
        "tf_name,tf_version,np_version",
        [
            ("tensorflow", "2.2.0", "1.18.0"),  # NumPy version not affected
            ("tensorflow-cpu", "2.5.0", "1.19.0"),  # a newer TF release
            ("tensorflow-gpu", "1.19", "2.0.0"),  # An older TF release
            ("intel-tensorflow", "2.1.0", "1.17.0"),  # Not affected
        ],
    )
    def _test_run_noop(self, context: Context, tf_name: str, tf_version: str, np_version: str) -> None:
        """Test wrong resolutions are not acceptable."""
        package_version = PackageVersion(
            name="numpy", version=f"=={np_version}", develop=False, index=Source("https://pypi.org/simple"),
        )

        tf_package_version = PackageVersion(
            name=tf_name, version=f"=={tf_version}", develop=False, index=Source("https://pypi.org/simple"),
        )

        state = State()
        state.add_resolved_dependency(tf_package_version.to_tuple())
        context.register_package_version(tf_package_version)

        unit = TensorFlow22NumPyStep()
        unit.pre_run()

        assert unit._message_logged is False

        with unit.assigned_context(context):
            assert unit.run(state, package_version) is None

        assert unit._message_logged is False



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Suggest not to use TensorFlow 2.1 with specific urllib3 versions that cause six import issues."""

import attr
from typing import Any
from typing import Optional
from typing import Tuple
from typing import List
from typing import Dict
from typing import TYPE_CHECKING
import logging

from thoth.common import get_justification_link as jl
from thoth.python import PackageVersion

from ...enums import RecommendationType
from ...exceptions import NotAcceptable
from ...state import State
from ...step import Step


if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext


_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class TensorFlow21Urllib3Step(Step):
    """A step that suggests not to use TensorFlow 2.1 with specific urllib3 versions that cause six import issues."""

    _message_logged = attr.ib(type=bool, default=False, init=False)

    # Run this step each time, regardless of when TensorFlow and urllib3 are resolved.
    MULTI_PACKAGE_RESOLUTIONS = True

    _MESSAGE = (
        "TensorFlow in version 2.1 can cause runtime errors when imported, caused by "
        "incompatibility between urllib3 and six packages"
    )
    _LINK = jl("tf_21_urllib3")
    _AFFECTED_URLLIB3_VERSIONS = frozenset({(1, 2), (1, 3), (1, 4), (1, 5)})

    def pre_run(self) -> None:
        """Initialize this pipeline unit before each run."""
        self._message_logged = False

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Optional[Dict[str, Any]]:
        """Register this pipeline unit for adviser if not using latest recommendations."""
        if (
            not builder_context.is_adviser_pipeline()
            or builder_context.recommendation_type == RecommendationType.LATEST
        ):
            return None

        if not builder_context.is_included(cls):
            return {}

        return None

    def run(
        self, state: State, package_version: PackageVersion
    ) -> Optional[Tuple[Optional[float], Optional[List[Dict[str, str]]]]]:
        """Suggest not to use TensorFlow 2.1 with specific urllib3 as issues with six were spotted on imports."""
        if (
            package_version.name != "urllib3"
            or package_version.semantic_version.release[:2] not in self._AFFECTED_URLLIB3_VERSIONS
        ):
            return None

        tensorflow_any = (
            state.resolved_dependencies.get("tensorflow")
            or state.resolved_dependencies.get("tensorflow-cpu")
            or state.resolved_dependencies.get("tensorflow-gpu")
            or state.resolved_dependencies.get("intel-tensorflow")
        )

        if not tensorflow_any or (tensorflow_any[1] != "2.1" and not tensorflow_any[1].startswith("2.1.")):
            return None

        if not self._message_logged:
            self._message_logged = True
            _LOGGER.warning("%s - see %s", self._MESSAGE, self._LINK)
            self.context.stack_info.append({"type": "WARNING", "message": self._MESSAGE, "link": self._LINK})

        raise NotAcceptable



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test using the right TensorFlow release for specific CUDA releases."""

from typing import Tuple
from itertools import product

import pytest

from thoth.adviser.enums import DecisionType
from thoth.adviser.enums import RecommendationType
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.sieves import TensorFlowCUDASieve
from thoth.adviser.context import Context
from thoth.python import PackageVersion
from thoth.python import Source

from ...base import AdviserTestCase


class TestTensorFlowCUDASieve(AdviserTestCase):
    """Test using the right TensorFlow release for specific CUDA releases."""

    def test_recommendation_types_considered(self) -> None:
        """Test recommendation types that were considered during implementation of this pipeline unit.

        This test will fail when a new recommendation type is added. In such case, review the pipeline unit
        implementation and adjust it accordingly.
        """
        assert set(RecommendationType) == {
            RecommendationType.STABLE,
            RecommendationType.TESTING,
            RecommendationType.LATEST,
            RecommendationType.PERFORMANCE,
            RecommendationType.SECURITY,
        }

    @pytest.mark.parametrize(
        "recommendation_type,cuda_version",
        list(
            product(
                [RecommendationType.STABLE, RecommendationType.PERFORMANCE, RecommendationType.SECURITY],
                ["8", "9", "10.0", "10.1"],
            )
        ),
    )
    def test_include(
        self, builder_context: PipelineBuilderContext, recommendation_type: RecommendationType, cuda_version: str
    ) -> None:
        """Test including this pipeline unit."""
        builder_context.decision_type = None
        builder_context.recommendation_type = recommendation_type
        builder_context.project.runtime_environment.cuda_version = cuda_version
        assert builder_context.is_adviser_pipeline()
        assert TensorFlowCUDASieve.should_include(builder_context) == {}

    @pytest.mark.parametrize(
        "recommendation_type,decision_type,cuda_version",
        [
            (RecommendationType.LATEST, None, "10.1"),  # Do not include for LATEST
            (RecommendationType.TESTING, None, "10.1"),  # Do not include for TESTING
            (None, DecisionType.RANDOM, "10.0"),  # A Dependency Monkey run
            (RecommendationType.STABLE, None, "UNKNOWN_CUDA_VERSION"),  # Unknown CUDA version
            (RecommendationType.STABLE, None, None),  # No CUDA version provided
        ],
    )
    def test_no_include(
        self,
        builder_context: PipelineBuilderContext,
        recommendation_type: RecommendationType,
        decision_type: DecisionType,
        cuda_version: str,
    ) -> None:
        """Test not including this pipeline unit."""
        builder_context.decision_type = decision_type
        builder_context.recommendation_type = recommendation_type
        builder_context.project.runtime_environment.cuda_version = cuda_version
        assert builder_context.is_adviser_pipeline() or builder_context.is_dependency_monkey_pipeline()
        assert TensorFlowCUDASieve.should_include(builder_context) is None

    @pytest.mark.parametrize(
        "cuda_version,expected_tf_1_support,expected_tf_2_support",
        [
            ("8", "_TF_1_CUDA_8_SUPPORT", "_EMPTY"),
            ("9", "_TF_1_CUDA_9_SUPPORT", "_EMPTY"),
            ("10.0", "_TF_1_CUDA_10_0_SUPPORT", "_TF_2_CUDA_10_0_SUPPORT"),
            ("10.1", "_EMPTY", "_TF_2_CUDA_10_1_SUPPORT"),
            ("UNKNOWN_CUDA_VERSION", "_EMPTY", "_EMPTY"),  # Should not happen based on should_include
        ],
    )
    def test_pre_run(
        self, context: Context, cuda_version: str, expected_tf_1_support: str, expected_tf_2_support: str
    ) -> None:
        """Test initializing the pipeline unit."""
        unit = TensorFlowCUDASieve()

        assert unit._tf_1_cuda_support is unit._EMPTY
        assert unit._tf_2_cuda_support is unit._EMPTY

        unit._messages_logged.add("foo")
        assert unit._messages_logged

        context.project.runtime_environment.cuda_version = cuda_version

        with unit.assigned_context(context):
            unit.pre_run()

        assert not unit._messages_logged
        assert unit._tf_1_cuda_support is getattr(unit, expected_tf_1_support)
        assert unit._tf_2_cuda_support is getattr(unit, expected_tf_2_support)

    @pytest.mark.parametrize("package_name", ["tensorflow", "tensorflow-gpu"])
    def test_unknown_tensorflow(self, context: Context, package_name: str) -> None:
        """Test not discarding if an unknown TensorFlow release is spotted."""
        context.project.runtime_environment.cuda_version = "10.0"
        package_version = PackageVersion(
            name=package_name, version="==42.30.03", develop=False, index=Source("https://pypi.org/simple"),
        )

        unit = TensorFlowCUDASieve()
        with unit.assigned_context(context):
            unit.pre_run()
            result = list(unit.run((pv for pv in (package_version,))))
            assert len(result) == 1
            assert result[0] is package_version, "The pipeline unit should keep the unknown TensorFlow release"

    @pytest.mark.parametrize(
        "package_name,package_version,cuda_version",
        [
            ("tensorflow", "2.3.0", "10.1"),
            ("tensorflow-gpu", "2.3.0", "10.1"),
            ("tensorflow", "2.2.0", "10.1"),
            ("tensorflow-gpu", "2.2.0", "10.1"),
            ("tensorflow", "2.1.0", "10.1"),
            ("tensorflow-gpu", "2.1.0", "10.1"),
            ("tensorflow", "2.0.0", "10.0"),
            ("tensorflow-gpu", "2.0.0", "10.0"),
            ("tensorflow_gpu", "1.15.0", "10.0"),
            ("tensorflow_gpu", "1.14.0", "10.0"),
            ("tensorflow_gpu", "1.13.1", "10.0"),
            ("tensorflow_gpu", "1.12.0", "9"),
            ("tensorflow_gpu", "1.11.0", "9"),
            ("tensorflow_gpu", "1.10.0", "9"),
            ("tensorflow_gpu", "1.9.0", "9"),
            ("tensorflow_gpu", "1.8.0", "9"),
            ("tensorflow_gpu", "1.7.0", "9"),
            ("tensorflow_gpu", "1.6.0", "9"),
            ("tensorflow_gpu", "1.5.0", "9"),
            ("tensorflow_gpu", "1.4.0", "8"),
            ("tensorflow_gpu", "1.3.0", "8"),
            ("tensorflow_gpu", "1.2.0", "8"),
            ("tensorflow_gpu", "1.1.0", "8"),
            ("tensorflow_gpu", "1.0.0", "8"),
        ],
    )
    def test_run_yield(self, context: Context, package_name: str, package_version: str, cuda_version: str) -> None:
        """Test packages the pipeline unit yields respecting CUDA version used.

        See the official docs for listing:
           https://www.tensorflow.org/install/source#gpu
        """
        context.project.runtime_environment.cuda_version = cuda_version
        package_version = PackageVersion(
            name=package_name, version=f"=={package_version}", develop=False, index=Source("https://pypi.org/simple"),
        )

        unit = TensorFlowCUDASieve()
        with unit.assigned_context(context):
            unit.pre_run()
            result = list(unit.run((pv for pv in (package_version,))))
            assert len(result) == 1
            assert result[0] is package_version

    @pytest.mark.parametrize(
        "pv,cuda_version",
        [
            *product((("tensorflow", "2.3.0"),), ("10.0", "9", "8")),
            *product((("tensorflow", "2.2.0"),), ("10.0", "9", "8")),
            *product((("tensorflow", "2.1.0"),), ("10.0", "9", "8")),
            *product((("tensorflow", "2.0.0"),), ("10.1", "9", "8")),
            *product((("tensorflow", "1.15.0"),), ("10.1", "10.0", "9", "8")),
            *product((("tensorflow-gpu", "1.15.0"),), ("10.1", "9", "8")),
            *product((("tensorflow-gpu", "1.14.0"),), ("10.1", "9", "8")),
            *product((("tensorflow-gpu", "1.13.1"),), ("10.1", "9", "8")),
            *product((("tensorflow-gpu", "1.12.0"),), ("10.1", "10.0", "8")),
            *product((("tensorflow-gpu", "1.11.0"),), ("10.1", "10.0", "8")),
            *product((("tensorflow-gpu", "1.10.0"),), ("10.1", "10.0", "8")),
            *product((("tensorflow-gpu", "1.9.0"),), ("10.1", "10.0", "8")),
            *product((("tensorflow-gpu", "1.8.0"),), ("10.1", "10.0", "8")),
            *product((("tensorflow-gpu", "1.7.0"),), ("10.1", "10.0", "8")),
            *product((("tensorflow-gpu", "1.6.0"),), ("10.1", "10.0", "8")),
            *product((("tensorflow-gpu", "1.5.0"),), ("10.1", "10.0", "8")),
            *product((("tensorflow-gpu", "1.4.0"),), ("10.1", "10.0", "9")),
            *product((("tensorflow-gpu", "1.3.0"),), ("10.1", "10.0", "9")),
            *product((("tensorflow-gpu", "1.2.0"),), ("10.1", "10.0", "9")),
            *product((("tensorflow-gpu", "1.1.0"),), ("10.1", "10.0", "9")),
            *product((("tensorflow-gpu", "1.0.0"),), ("10.1", "10.0", "9")),
        ],
    )
    def test_run_no_yield(self, context: Context, pv: Tuple[str, str], cuda_version: str) -> None:
        """Test discarding packages that do not conform to the support matrix.

        See the official docs for listing:
           https://www.tensorflow.org/install/source#gpu
        """
        context.project.runtime_environment.cuda_version = cuda_version
        package_version = PackageVersion(
            name=pv[0], version=f"=={pv[1]}", develop=False, index=Source("https://pypi.org/simple"),
        )

        unit = TensorFlowCUDASieve()
        with unit.assigned_context(context):
            unit.pre_run()
            result = list(unit.run((pv for pv in (package_version,))))
            assert len(result) == 0

    def test_sieve_multiple(self, context: Context) -> None:
        """Test proper implementation of the filtering mechanism."""
        context.project.runtime_environment.cuda_version = "10.0"
        source = Source("https://pypi.org/simple")
        pv_1 = PackageVersion(name="tensorflow-gpu", version="==1.12.0", develop=False, index=source,)
        pv_2 = PackageVersion(name="tensorflow", version="==2.0.0", develop=False, index=source,)
        pv_3 = PackageVersion(name="tensorflow", version="==1.13.0", develop=False, index=source,)

        unit = TensorFlowCUDASieve()
        with unit.assigned_context(context):
            unit.pre_run()
            result = list(unit.run((pv for pv in (pv_1, pv_2, pv_3))))
            assert len(result) == 1
            assert result[0] == pv_2



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A sieve for filtering out disabled Python package indexes."""

import logging
from typing import Any
from typing import Optional
from typing import Dict
from typing import Generator
from typing import TYPE_CHECKING

import attr
from thoth.python import PackageVersion
from thoth.storages.exceptions import NotFoundError

from ..sieve import Sieve

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class PackageIndexSieve(Sieve):
    """Filter out disabled Python package indexes."""

    _cached_records: Dict[str, Optional[bool]] = attr.ib(default=attr.Factory(dict), kw_only=True)

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Optional[Dict[str, Any]]:
        """Remove indexes which are not enabled in pipeline configuration."""
        if not builder_context.is_included(cls):
            return {}

        return None

    def run(self, package_versions: Generator[PackageVersion, None, None]) -> Generator[PackageVersion, None, None]:
        """Filter out package versions based on disabled Python package index."""
        for package_version in package_versions:
            if package_version.index.url in self._cached_records:
                is_enabled = self._cached_records[package_version.index.url]
            else:
                try:
                    is_enabled = self.context.graph.is_python_package_index_enabled(package_version.index.url)
                except NotFoundError:
                    # A special value of None marks non-existing index in thoth's knowledge base.
                    is_enabled = None

            self._cached_records[package_version.index.url] = is_enabled
            if is_enabled is None:
                _LOGGER.debug(
                    "Removing Python package version %r as used index is not registered", package_version.to_tuple(),
                )
                continue

            if not is_enabled:
                _LOGGER.debug(
                    "Removing Python package version %r as used index is not enabled", package_version.to_tuple(),
                )
                continue

            yield package_version



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A boot to adjust RHEL version to its major version."""

import logging
from typing import Optional
from typing import Dict
from typing import Any
from typing import TYPE_CHECKING

import attr

from thoth.common import get_justification_link as jl

from ..boot import Boot

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class RHELVersionBoot(Boot):
    """A boot that changes version of RHEL used.

    RHEL guarantees ABI compatibility across major minor releases.
    """

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Optional[Dict[str, Any]]:
        """Register self, always."""
        if not builder_context.is_included(cls):
            return {}

        return None

    def run(self) -> None:
        """Discard any minor release in RHEL."""
        os_name = self.context.project.runtime_environment.operating_system.name
        os_version = self.context.project.runtime_environment.operating_system.version
        if os_name == "rhel" and os_version is not None:
            version_parts = os_version.split(".", maxsplit=1)
            if len(version_parts) > 1:
                _LOGGER.info(
                    "RHEL major releases guarantee ABI compatibility across minor releases; "
                    "discarding minor release information and using RHEL version %r - see %s",
                    version_parts[0],
                    jl("rhel_version"),
                )
                self.context.project.runtime_environment.operating_system.version = version_parts[0]



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Implementation of a predictor used for generating combinations of packages faster."""

from typing import Tuple
from typing import Set
import logging

import attr

from ..predictor import Predictor
from ..state import State

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class PackageCombinations(Predictor):
    """A predictor used for generating combinations of packages faster."""

    package_combinations = attr.ib(type=Set[str], default=attr.Factory(set), kw_only=True, converter=set)

    def pre_run(self) -> None:
        """Check attributes set up."""
        if not self.package_combinations:
            raise ValueError("No package combinations supplied to the predictor")

    def run(self) -> Tuple[State, Tuple[str, str, str]]:
        """Run the predictor."""
        state = self.context.beam.get_last()
        if state is None:
            state = self.context.beam.get_random()

        # Expand while we do not have just all the package combinations to generate.
        for unresolved_dependency in state.unresolved_dependencies.keys():
            if unresolved_dependency not in self.package_combinations:
                return state, state.get_random_unresolved_dependency(unresolved_dependency)

        return state, state.get_random_unresolved_dependency()



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Filter out software stacks allowing the pipeline to produce just one software stack with specific package.

This stride will allow the pipeline to produce a stack with only a specific package-version just once - e.g. if
two or more software stacks could be resolved with tensorflow==2.0.0, this stride will allow the
pipeline to produce just one software stack with tensorflow==2.0.0, other software stacks will be filtered out.
"""

import logging
from typing import Any
from typing import Dict
from typing import Optional
from typing import TYPE_CHECKING

import attr

from ..state import State
from ..stride import Stride
from ..exceptions import StrideError
from ..exceptions import NotAcceptable

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class OneVersionStride(Stride):
    """Filter out software stacks allowing the pipeline to produce just one software stack with specific package."""

    version_seen = attr.ib(type=Optional[str], default=None)  # type: ignore

    CONFIGURATION_DEFAULT = {
        "package_name": None,
        "only_once": True,
    }

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Optional[Dict[str, Any]]:
        """Include this pipeline unit only if user asks for it explicitly."""
        return None

    def pre_run(self) -> None:
        """Initialize internal state of the unit."""
        self.version_seen = None

        if self.configuration["package_name"] is None:
            raise StrideError(f"No package name provided to the {self.__class__.__name__!r} pipeline unit")

    def run(self, state: State) -> None:
        """Filter out software stacks, allow only a package with specific version being produced just once."""
        package_tuple = state.resolved_dependencies.get(self.configuration["package_name"])
        if package_tuple is None:
            return

        if self.version_seen is None:
            # First time seen, mark and continue.
            self.version_seen = package_tuple[1]
            return

        if self.version_seen == package_tuple[1] and not self.configuration["only_once"]:
            return

        if self.version_seen == package_tuple[1]:
            raise NotAcceptable(
                f"Package {package_tuple[0]!r} in version {package_tuple[1]} was "
                "already seen and only_once flag was set"
            )

        raise NotAcceptable(
            f"Package {package_tuple[0]!r} was already seen in version {self.version_seen!r}, "
            f"not accepting stack that has the given package in version {package_tuple[1]!r}"
        )



"""Setup configuration for adviser module."""

import os
import sys
from setuptools import setup
from setuptools.command.test import test as TestCommand  # noqa


def get_install_requires():
    """Get requirements for adviser module."""
    with open("requirements.txt", "r") as requirements_file:
        res = requirements_file.readlines()
        return [req.split(" ", maxsplit=1)[0] for req in res if req]


def get_version():
    """Get current version of adviser module."""
    with open(os.path.join("thoth", "adviser", "__init__.py")) as f:
        content = f.readlines()

    for line in content:
        if line.startswith("__version__ ="):
            # dirty, remove trailing and leading chars
            return line.split(" = ")[1][1:-2]
    raise ValueError("No version identifier found")


def read(fname):
    """Read."""
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


class Test(TestCommand):
    """Introduce test command to run testsuite using pytest."""

    _IMPLICIT_PYTEST_ARGS = [
        "--timeout=60",
        "--cov=./thoth",
        "--mypy",
        "--capture=no",
        "--verbose",
        "-l",
        "-s",
        "-vv",
        "--hypothesis-show-statistics",
        "tests/",
    ]

    user_options = [("pytest-args=", "a", "Arguments to pass into py.test")]

    def initialize_options(self):
        """Initialize cli options."""
        super().initialize_options()
        self.pytest_args = None

    def finalize_options(self):
        """Finalize cli options."""
        super().finalize_options()
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        """Run module tests."""
        import pytest

        passed_args = list(self._IMPLICIT_PYTEST_ARGS)

        if self.pytest_args:
            self.pytest_args = [arg for arg in self.pytest_args.split() if arg]
            passed_args.extend(self.pytest_args)

        sys.exit(pytest.main(passed_args))


VERSION = get_version()
setup(
    name="thoth-adviser",
    version=VERSION,
    description="Package and package stack adviser for the Thoth project",
    long_description=read("README.rst"),
    author="Fridolin Pokorny",
    author_email="fridolin@redhat.com",
    license="GPLv3+",
    packages=[
        "thoth.adviser",
        "thoth.adviser.boots",
        "thoth.adviser.boots._debug",
        "thoth.adviser.predictors",
        "thoth.adviser.sieves",
        "thoth.adviser.sieves.backports",
        "thoth.adviser.sieves.tensorflow",
        "thoth.adviser.sieves.pandas",
        "thoth.adviser.sieves.setuptools",
        "thoth.adviser.steps",
        "thoth.adviser.steps",
        "thoth.adviser.steps._debug",
        "thoth.adviser.steps.tensorflow",
        "thoth.adviser.strides",
        "thoth.adviser.wraps",
        "thoth.adviser.wraps.python",
        "thoth.adviser.wraps.pytorch",
        "thoth.adviser.wraps.tensorflow",
    ],
    url="https://github.com/thoth-station/adviser",
    download_url="https://pypi.org/project/thoth-adviser",
    package_data={"thoth.adviser": ["py.typed"]},
    entry_points={"console_scripts": ["thoth-adviser=thoth.adviser.cli:cli"]},
    zip_safe=False,
    install_requires=get_install_requires(),
    cmdclass={"test": Test},
    long_description_content_type="text/x-rst",
    command_options={"build_sphinx": {"version": ("setup.py", VERSION), "release": ("setup.py", VERSION),}},
)



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Implementation of steps used during resolution."""

from .aicoe import AICoEReleasesStep
from .cve import CvePenalizationStep
from .dropout import DropoutStep
from .security_indicators import SecurityIndicatorStep
from .tensorflow import TensorFlow21Urllib3Step
from .tensorflow import TensorFlow22ProbabilityStep
from .tensorflow import TensorFlowAVX2Step
from .tensorflow import TensorFlow113NumPyStep
from .tensorflow import TensorFlow114GastStep
from .tensorflow import TensorFlow22NumPyStep
from .tensorflow import TensorFlowRemoveSciPyStep

# from ._debug import MockScoreStep


# Relative ordering of units is relevant, as the order specifies order
# in which the asked to be registered - any dependencies between them
# can be mentioned here.
__all__ = [
    "AICoEReleasesStep",
    "CvePenalizationStep",
    "DropoutStep",
    "SecurityIndicatorStep",
    # "MockScoreStep",
    "TensorFlow21Urllib3Step",
    "TensorFlow22ProbabilityStep",
    "TensorFlowAVX2Step",
    "TensorFlow113NumPyStep",
    "TensorFlow114GastStep",
    "TensorFlow22NumPyStep",
    "TensorFlowRemoveSciPyStep",
]



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Helpers for the testsuite and adviser implementation."""

import sys
import functools
from typing import Any

import tests.units as units
import thoth.adviser


def use_test_units(func: Any) -> Any:
    """Substitute implemented units in adviser implementation with the ones provided by testsuite."""
    # noqa
    @functools.wraps(func)
    def wrapped(*args: Any, **kwargs: Any) -> Any:
        """Substitute implemented units with the testing ones."""
        sys.modules["thoth.adviser"].boots = units.boots
        sys.modules["thoth.adviser.boots"] = units.boots
        sys.modules["thoth.adviser"].sieves = units.sieves
        sys.modules["thoth.adviser.sieves"] = units.sieves
        sys.modules["thoth.adviser"].steps = units.steps
        sys.modules["thoth.adviser.steps"] = units.steps
        sys.modules["thoth.adviser"].strides = units.strides
        sys.modules["thoth.adviser.strides"] = units.strides
        sys.modules["thoth.adviser"].wraps = units.wraps
        sys.modules["thoth.adviser.wraps"] = units.wraps
        try:
            return func(*args, **kwargs)
        finally:
            sys.modules["thoth.adviser"].boots = thoth.adviser.boots
            sys.modules["thoth.adviser.boots"] = thoth.adviser.boots
            sys.modules["thoth.adviser"].sieves = thoth.adviser.sieves
            sys.modules["thoth.adviser.sieves"] = thoth.adviser.sieves
            sys.modules["thoth.adviser"].steps = thoth.adviser.steps
            sys.modules["thoth.adviser.steps"] = thoth.adviser.steps
            sys.modules["thoth.adviser"].strides = thoth.adviser.strides
            sys.modules["thoth.adviser.strides"] = thoth.adviser.strides
            sys.modules["thoth.adviser"].wraps = thoth.adviser.wraps
            sys.modules["thoth.adviser.wraps"] = thoth.adviser.wraps

    return wrapped



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A generic unit representatives used for testing.."""

from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import TYPE_CHECKING

from thoth.adviser.stride import Stride
from thoth.adviser.state import State

if TYPE_CHECKING:
    from thoth.adviser.pipeline_builder import PipelineBuilderContext


class Stride1(Stride):
    """A testing stride implementation."""

    CONFIGURATION_DEFAULT = {"linus": {"residence": "oregon", "children": 3, "parents": ["nils", "anna"]}}

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Optional[Dict[str, Any]]:
        """Check if this pipeline unit should be included in the pipeline configuration."""

    def run(self, state: State) -> Optional[Tuple[float, List[Dict[str, str]]]]:
        """Run noop method."""


class Stride2(Stride):
    """A testing stride implementation."""

    CONFIGURATION_DEFAULT = {"foo": None}

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Optional[Dict[str, Any]]:
        """Check if this pipeline unit should be included in the pipeline configuration."""

    def run(self, state: State) -> Optional[Tuple[float, List[Dict[str, str]]]]:
        """Run noop method."""


__all__ = ["Stride1", "Stride2"]



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A sieve to remove backport of functools intended for older Python versions."""

import logging
from typing import Any
from typing import Dict
from typing import Generator
from typing import Optional
from typing import TYPE_CHECKING

import attr
from thoth.common import get_justification_link as jl
from thoth.python import PackageVersion

from ...exceptions import SkipPackage
from ...sieve import Sieve

if TYPE_CHECKING:
    from ...pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class Functools32BackportSieve(Sieve):
    """Remove backport of functools, available in the standard library starting Python 3.2.

    https://pypi.org/project/functools32/
    https://docs.python.org/3/library/functools.html
    """

    _MESSAGE = (
        f"Dependency 'functools32' removed: functools is available in Python standard "
        f"library starting Python 3.2 - {jl('backports')}"
    )

    _logged = attr.ib(default=False, type=bool, init=False)

    def pre_run(self) -> None:
        """Initialize self before running."""
        self._logged = False

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Optional[Dict[str, Any]]:
        """Include for Python 3.3 and above for adviser and dependency monkey runs."""
        if builder_context.is_included(cls) or builder_context.project.runtime_environment.python_version is None:
            return None

        if builder_context.project.runtime_environment.get_python_version_tuple() >= (3, 2):
            return {}

        return None

    def run(self, package_versions: Generator[PackageVersion, None, None]) -> Generator[PackageVersion, None, None]:
        """Remove dependency functools32 for newer Python versions."""
        for package_version in package_versions:
            if package_version.name == "functools32":
                if not self._logged:
                    self.context.stack_info.append({"type": "WARNING", "message": self._MESSAGE})
                    _LOGGER.warning(self._MESSAGE)
                    self._logged = True

                raise SkipPackage

            yield package_version



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test a step for TensorFlow 1.13.1 and its compatibility with NumPy>=1.16.0."""

from itertools import product

import pytest

from thoth.adviser.context import Context
from thoth.adviser.enums import DecisionType
from thoth.adviser.enums import RecommendationType
from thoth.adviser.exceptions import NotAcceptable
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.state import State
from thoth.adviser.steps import TensorFlow113NumPyStep
from thoth.python import PackageVersion
from thoth.python import Source

from ...base import AdviserTestCase


class TestTensorFlow113NumPyStep(AdviserTestCase):
    """TensorFlow 1.13.1 is compatible with NumPy>=1.16.0, not with >=1.13.3 as stated in setup.py."""

    @pytest.mark.parametrize(
        "recommendation_type",
        [
            RecommendationType.STABLE,
            RecommendationType.TESTING,
            RecommendationType.PERFORMANCE,
            RecommendationType.STABLE,
        ],
    )
    def test_include(self, builder_context: PipelineBuilderContext, recommendation_type: RecommendationType) -> None:
        """Test including this pipeline unit."""
        builder_context.decision_type = None
        builder_context.recommendation_type = recommendation_type
        assert builder_context.is_adviser_pipeline()
        assert TensorFlow113NumPyStep.should_include(builder_context) == {}

    @pytest.mark.parametrize(
        "recommendation_type,decision_type",
        [(RecommendationType.LATEST, None), (None, DecisionType.RANDOM),],  # A Dependency Monkey run.
    )
    def test_no_include(
        self, builder_context: PipelineBuilderContext, recommendation_type, decision_type: DecisionType,
    ) -> None:
        """Test not including this pipeline unit step."""
        builder_context.decision_type = decision_type
        builder_context.recommendation_type = recommendation_type
        assert TensorFlow113NumPyStep.should_include(builder_context) is None

    @pytest.mark.parametrize(
        "tf_name,np_version",
        list(product(["tensorflow", "tensorflow-cpu", "tensorflow-gpu"], ["1.13.3", "1.14.0", "1.15.0", "1.15.5"])),
    )
    def test_run_not_acceptable(self, context: Context, tf_name: str, np_version: str) -> None:
        """Test wrong resolutions are not acceptable."""
        package_version = PackageVersion(
            name="numpy", version=f"=={np_version}", develop=False, index=Source("https://pypi.org/simple"),
        )

        state = State()
        state.add_resolved_dependency((tf_name, "1.13.1", "https://pypi.org/simple"))

        unit = TensorFlow113NumPyStep()
        unit.pre_run()

        assert not context.stack_info

        with unit.assigned_context(context):
            with pytest.raises(NotAcceptable):
                assert unit._message_logged is False
                unit.run(state, package_version)

        assert unit._message_logged is True
        assert context.stack_info
        assert self.verify_justification_schema(context.stack_info) is True

    @pytest.mark.parametrize(
        "tf_name,np_version",
        list(product(["tensorflow", "tensorflow-cpu", "tensorflow-gpu"], ["1.16.0", "1.17.0", "1.19.0"])),
    )
    def test_run_noop(self, tf_name: str, np_version: str) -> None:
        """Test wrong resolutions are not acceptable."""
        package_version = PackageVersion(
            name="numpy", version=f"=={np_version}", develop=False, index=Source("https://pypi.org/simple"),
        )

        state = State()
        state.add_resolved_dependency((tf_name, "1.13.1", "https://pypi.org/simple"))

        unit = TensorFlow113NumPyStep()

        assert unit._message_logged is False
        assert unit.run(state, package_version) is None
        assert unit._message_logged is False
